{ File   : DATExt.dpr
  Author : Deniz Oezmen
  Created: 2007-06-27
  Changed: 2007-06-27

  Arcanum DAT extractor.
}
program DATExt;

{$APPTYPE CONSOLE}

uses
  SysUtils, ZLib;

type
  TDATEntry   = record
                  NameLen   : Longword;
                  Name      : string;
                  Unknown1,
                  EntryType,
                  UncompSize,
                  CompSize,
                  Offset    : Longword;
                end;

  TDATDir     = record
                  NumEntries : Longword;
                  Entries    : array of TDATEntry;
                end;

  TDATTrailer = packed record
                  Unknown    : array[0..23] of Byte;
                  TrailerSize: Longword;
                end;

  TBuffer     = array of Byte;

var
  DATFile: file;

function OpenDATFile(var DATFile: file; DATName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(DATFile, DATName);

  {$I-}
  Reset(DATFile, 1);
  {$I+}
  OpenDATFile := (IOResult = 0)
end;

function ReadTrailer(out DATTrailer: TDATTrailer): Boolean;
var
  BytesRead: Longword;
begin
  Seek(DATFile, FileSize(DATFile) - SizeOf(TDATTrailer));
  BlockRead(DATFile, DATTrailer, SizeOf(TDATTrailer), BytesRead);
  Result := (BytesRead = SizeOf(TDATTrailer))
end;

function Decompress(const InBuffer: TBuffer; var OutBuffer: TBuffer): Boolean;
var
  DecompSize: Longint;
  TempBuffer: Pointer;
begin
  DecompressBuf(@InBuffer[0], Length(InBuffer), Length(OutBuffer), TempBuffer,
    DecompSize);

  Result := (DecompSize = Length(OutBuffer));

  if Result then
    Move(TempBuffer^, OutBuffer[0], DecompSize);

  FreeMem(TempBuffer, DecompSize)
end;

procedure ExtractFiles(const DATTrailer: TDATTrailer);
var
  Dir      : TDATDir;
  i        : Longint;
  OutFile  : file;
  InBuffer,
  OutBuffer: TBuffer;
begin
  {$I-}
  Seek(DATFile, Longword(FileSize(DATFile)) - DATTrailer.TrailerSize);
  BlockRead(DATFile, Dir.NumEntries, SizeOf(Dir.NumEntries));

  SetLength(Dir.Entries, Dir.NumEntries);

  for i := 0 to High(Dir.Entries) do
    with Dir.Entries[i] do
    begin
      BlockRead(DATFile, NameLen, SizeOf(NameLen));
      SetLength(Name, NameLen);
      BlockRead(DATFile, Name[1], NameLen);
      // remove zero terminator(s)
      Name := string(PChar(Name));
      BlockRead(DATFile, Unknown1, SizeOf(Unknown1));
      BlockRead(DATFile, EntryType, SizeOf(EntryType));
      BlockRead(DATFile, UncompSize, SizeOf(UncompSize));
      BlockRead(DATFile, CompSize, SizeOf(CompSize));
      BlockRead(DATFile, Offset, SizeOf(Offset));
    end;

  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  error reading structures');
    Finalize(Dir.Entries);
    Exit
  end;

  for i := 0 to High(Dir.Entries) do
  begin
    with Dir.Entries[i] do
    begin
      WriteLn(Format('  %s', [Name]));

      case EntryType of
        $00000001:
          begin
            Seek(DATFile, Offset);
            SetLength(OutBuffer, UncompSize);
            BlockRead(DATFile, OutBuffer[0], Length(OutBuffer))
          end;
        $00000002:
          begin
            Seek(DATFile, Offset);
            SetLength(InBuffer, CompSize);
            SetLength(OutBuffer, UncompSize);
            BlockRead(DATFile, InBuffer[0], Length(InBuffer));
            if not Decompress(InBuffer, OutBuffer) then
              WriteLn('    decompression error')
          end;
        $00000400:
          begin
            if not ForceDirectories(ExpandFileName(Name)) then
              WriteLn('    directory creation error');
            Continue
          end
      else
        WriteLn(Format('    unknown type %.8x', [EntryType]));
        Continue
      end;

      FileMode := fmOpenWrite;
      AssignFile(OutFile, Name);
      {$I-}
      Rewrite(OutFile, 1);
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    file creation error');
        Continue
      end;

      {$I-}
      BlockWrite(OutFile, OutBuffer[0], Length(OutBuffer));
      {$I+}
      if IOResult <> 0 then
        WriteLn('    write error')
    end
  end;

  Finalize(InBuffer);
  Finalize(OutBuffer);
  Finalize(Dir.Entries)
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i         : Byte;
  FileName,
  Dir,
  DATName   : string;
  DATTrailer: TDATTrailer;
  SearchRec : TSearchRec;
  PFilePara : PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara 
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      DATName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenDATFile(DATFile, DATName) then
        if ReadTrailer(DATTrailer) then
        begin
          Writeln('Extracting from ', DATName, ' ...');
          ExtractFiles(DATTrailer);
          Close(DATFile)
        end
        else
          Writeln(DATName, ' is not a valid DAT file.')
      else
        Writeln(DATName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found')
end;

begin
  if ParamCount = 1 then
    ParseFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.
